class ByError(Exception):
    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(message)


class SurahNotFound(Exception):
    def __init__(self, *args):
        super().__init__(*args)


class VersetNotFound(Exception):
    def __init__(self, *args):
        super().__init__(*args)


class FormatValueGet(Exception):
    def __init__(self, *args):
        super().__init__(*args)



from typing import TypedDict, List, Literal
from enum import Enum


class QuranVersetData(TypedDict):
    """
    Représente les informations d'un verset du Coran.

    Attributes:
        position (int): Position globale du verset dans le Coran.
        text (str): Le texte du verset en translittération.
        position_ds_sourate (int): Position du verset dans la sourate.
        juz (int): Le numéro du Juz auquel appartient le verset.
        manzil (int): Le numéro du Manzil auquel appartient le verset.
        page (int): Le numéro de page où se trouve le verset.
        ruku (int): Le numéro du Ruku auquel appartient le verset.
        hizbQuarter (int): Le quart de Hizb auquel appartient le verset.
        sajda (bool): Indique si le verset contient une prosternation (Sajda).
        text_arabe (str): Le texte du verset en arabe.
    """
    position: int
    text: str
    position_ds_sourate: int
    juz: int
    manzil: int
    page: int
    ruku: int
    hizbQuarter: int
    sajda: bool
    text_arabe: str


class QuranSourateData(TypedDict):
    """
    Représente les informations d'une sourate du Coran.

    Attributes:
        position (int): La position de la sourate dans le Coran.
        nom (str): Le nom de la sourate en arabe.
        nom_phonetique (str): La translittération du nom de la sourate.
        englishNameTranslation (str): La traduction anglaise du nom de la sourate.
        revelation (str): Le type de révélation (Medinois ou Mecquoise).
        versets (List[QuranVersetData]): La liste des versets de la sourate.
        nom_sourate (str): Le nom de la sourate en français.
    """
    position: int
    nom: str
    nom_phonetique: str
    englishNameTranslation: str
    revelation: str
    versets: List[QuranVersetData]
    nom_sourate: str


class QuranData(TypedDict):
    """
    Représente les données du Coran avec toutes les sourates.

    Attributes:
        sourates (List[QuranSouratetData]): Liste des sourates avec leurs informations.
    """
    sourates: List[QuranSourateData]


class RevelationType(Enum):
    """
    Enumération des types de révélations des sourates.

    Values:
        MEDINOIS (str): Révélée à Médine.
        MECQUOISE (str): Révélée à La Mecque.
    """
    MEDINOIS = "Medinois"
    MECQUOISE = "Mecquoise"


class TypedMetaSurah(TypedDict):
    position: int
    name_arabic: str
    name_phonetic: str
    name_english: str
    name_french: str
    revelation_type: str
    number_of_ayahs: int
    sajdas_recommended: bool
    sajdas_obligatory: bool
    url_quran: str

ListMetaSurahs = List[TypedMetaSurah]


FilterVersets = Literal['position_ds_sourate']


class FilterVersetsDict (TypedDict):
    position_ds_sourate: int



import json
import pathlib as plib
from .interface import QuranData, QuranSourateData, RevelationType
from .services import Services
from typing import Optional, Dict, List, Literal, Union


class QuranServices:
    def __init__(self, path_database: str) -> None:
        self.path_database = path_database
        self.__services = Services()
        self.__response = self._request_quran_json()
        self.__save()

    def _request_quran_json(self) -> dict:
        return self.__services.quran_to_json()

    def __save(self) -> None:
        with open(plib.Path(self.path_database).joinpath("quran.json"), "w") as file:
            file.write(json.dumps(self.__response, indent=4))
        return None

    @property
    def resp(self) -> QuranData:
        return self.__response


class QuranReaderBase:
    """
    Classe permettant de lire et d'interroger les données du Coran depuis un fichier JSON.

    Attributes:
        path_database (str): Le chemin vers le dossier contenant les données du Coran.
        path_quran (Path): Le chemin complet vers le fichier quran.json.
    """

    def __init__(self, path_database: str) -> None:
        """
        Initialise le lecteur du Coran avec le chemin vers la base de données.

        Args:
            path_database (str): Chemin vers le dossier de la base de données.
        """
        self.path_database = path_database
        self.path_quran = plib.Path(path_database).joinpath("quran.json")

        if not self.path_quran.exists():
            quran_services = QuranServices(path_database)
            self.__quran = quran_services.resp
        else:
            self.__quran = self.__read_init()

        self.__sourate_by_number: Optional[Dict[int, QuranSourateData]] = None
        self.__sourate_by_name_fr: Optional[Dict[str, QuranSourateData]] = None
        self.__sourate_by_name_arabic: Optional[Dict[str,
                                                     QuranSourateData]] = None

    @staticmethod
    def revelation_type_enum() -> RevelationType:
        """
        Retourne l'énumération des types de révélation (Medinois ou Mecquoise).

        Returns:
            RevelationType: L'énumération des types de révélation.
        """
        return RevelationType

    def __read_init(self) -> QuranData:
        """
        Lit le fichier quran.json et initialise les données du Coran.

        Returns:
            QuranData: Les données du Coran sous forme de dictionnaire.

        Raises:
            Exception: En cas d'erreur lors de la lecture du fichier quran.json.
        """
        try:
            with open(self.path_quran, "r") as file:
                buffer = file.read()
                parser_json = json.loads(buffer)
                return parser_json
        except Exception as e:
            raise Exception(f'Error reading file quran.json {str(e)}')

    @property
    def quran(self) -> QuranData:
        """
        Retourne les données complètes du Coran.

        Returns:
            QuranData: Les données du Coran.
        """
        return self.__quran

    @property
    def sourate_by_number(self) -> Dict[int, QuranSourateData]:
        """
        Retourne un dictionnaire des sourates indexées par leur numéro de position.

        Returns:
            Dict[int, QuranSouratetData]: Les sourates indexées par position.
        """
        if self.__sourate_by_number is None:
            self.__sourate_by_number = {
                s['position']: s for s in self.quran['sourates']}
        return self.__sourate_by_number

    @property
    def sourate_by_name_fr(self) -> Dict[str, QuranSourateData]:
        """
        Retourne un dictionnaire des sourates indexées par leur nom en français.

        Returns:
            Dict[str, QuranSouratetData]: Les sourates indexées par nom en français.
        """
        if self.__sourate_by_name_fr is None:
            self.__sourate_by_name_fr = {
                s['nom_sourate']: s for s in self.quran['sourates']}
        return self.__sourate_by_name_fr

    @property
    def sourate_by_name_arabic(self) -> Dict[str, QuranSourateData]:
        """
        Retourne un dictionnaire des sourates indexées par leur nom en arabe.

        Returns:
            Dict[str, QuranSouratetData]: Les sourates indexées par nom en arabe.
        """
        if self.__sourate_by_name_arabic is None:
            self.__sourate_by_name_arabic = {
                s['nom']: s for s in self.quran['sourates']}
        return self.__sourate_by_name_arabic

    def search_by_number(self, number: int) -> Optional[QuranSourateData]:
        """
        Recherche une sourate par son numéro de position.

        Args:
            number (int): Le numéro de la sourate.

        Returns:
            Optional[QuranSouratetData]: Les informations de la sourate, ou None si non trouvée.
        """
        return self.sourate_by_number.get(number, None)

    def search_by_name_fr(self, namefr: str) -> Optional[QuranSourateData]:
        """
        Recherche une sourate par son nom en français.

        Args:
            namefr (str): Le nom de la sourate en français.

        Returns:
            Optional[QuranSouratetData]: Les informations de la sourate, ou None si non trouvée.
        """
        return self.sourate_by_name_fr.get(namefr, None)

    def search_by_name_arabic(self, namearabic: str) -> Optional[QuranSourateData]:
        """
        Recherche une sourate par son nom en arabe.

        Args:
            namearabic (str): Le nom de la sourate en arabe.

        Returns:
            Optional[QuranSouratetData]: Les informations de la sourate, ou None si non trouvée.
        """
        return self.sourate_by_name_arabic.get(namearabic, None)

    def get_all_by_type_revelation(self, type_of_revelation: Union[RevelationType, Literal["Medinois", "Mecquoise"]]) -> List[QuranSourateData]:
        """
        Retourne toutes les sourates correspondant au type de révélation spécifié.

        Args:
            type_of_revelation (Union[RevelationType, Literal["Medinois", "Mecquoise"]]): Le type de révélation (Medinois ou Mecquoise).

        Returns:
            List[QuranSouratetData]: Liste des sourates correspondant au type de révélation.
        """
        if isinstance(type_of_revelation, RevelationType):
            return [sourate for sourate in self.quran['sourates'] if sourate['revelation'] == type_of_revelation.value]
        else:
            return [sourate for sourate in self.quran['sourates'] if sourate['revelation'] == type_of_revelation]



from .metasurahs import MetaSurahs
from .meta_quran_reader import MetaQuranReader
from .yt_services import YoutubeServices
from .memory_quran import MemoryQuran
from .config import Config
from pathlib import Path


class MyDeen:
    """
    Point d'entrée principal du package `mydeen`.
    Fournit l'accès aux différents services liés au Coran, YouTube et mémorisation.
    """

    def __init__(self):
        path = Path(__file__).parent / "data"
        path.mkdir(parents=True, exist_ok=True)
        self.path_database = path.as_posix()
        self.setup_all()

    def config_url(self) -> Config:
        """Retourne la configuration des URLs de ressources utilisées."""
        return Config()

    def meta_surahs(self) -> MetaSurahs:
        """Retourne les métadonnées des sourates du Coran."""
        return MetaSurahs(self.path_database)

    def memory_quran(self) -> MemoryQuran:
        """Retourne l'outil de gestion des parties du Coran à mémoriser."""
        return MemoryQuran()

    def quran_reader(self) -> MetaQuranReader:
        """Retourne un lecteur du texte du Coran avec métadonnées."""
        return MetaQuranReader(self.path_database)

    def yt_services(self, api_key: str) -> YoutubeServices:
        """Retourne un service de récupération des vidéos YouTube."""
        return YoutubeServices(api_key)

    def setup_all(self) -> None:
        """Initialise les fichiers et les données nécessaires (cache, CSV, etc.)."""
        self.meta_surahs()
        self.quran_reader()



from __future__ import annotations

from mydeen.mydeen import MyDeen
from mydeen.config import Config
from mydeen.exception_error import (
    SurahNotFound,
    VersetNotFound,
    FormatValueGet,
    ByError,
)
from mydeen.interface import (
    QuranSourateData,
    QuranVersetData,
    QuranData,
    RevelationType,
    TypedMetaSurah,
    ListMetaSurahs,
)
from mydeen.yt_services import Channels
from mydeen.memory_quran import (
    MemoryQuran,
    MemoryParts,
    MemoryQuranData,
    PartsMemoryQuranData,
    PartsNameEnum,
)

__all__ = [
    "MyDeen",
    "Config",
    "SurahNotFound",
    "VersetNotFound",
    "FormatValueGet",
    "ByError",
    "QuranSourateData",
    "QuranVersetData",
    "QuranData",
    "RevelationType",
    "TypedMetaSurah",
    "ListMetaSurahs",
    "Channels",
    "MemoryQuran",
    "MemoryParts",
    "MemoryQuranData",
    "PartsMemoryQuranData",
    "PartsNameEnum",
]



from typing import NamedTuple, TypedDict, Dict, List
from enum import Enum
from .metasurahs import MetaSurahs, ListMetaSurahs
import pandas as pd


class MemoryQuranData(NamedTuple):
    """
    Représente une plage de sourates du Coran à mémoriser.

    Attributs :
        start (int) : Numéro de la sourate de début.
        end (int) : Numéro de la sourate de fin.
    """
    start: int
    end: int


class PartsNameEnum(str, Enum):
    as_sab_u_t_tiwal = "as_sab_u_t_tiwal"
    al_miin = "al_miin"
    al_mathani = "al_mathani"
    al_mufassal = "al_mufassal"
    mufassal_tiwal = "mufassal_tiwal"
    mufassal_awsat = "mufassal_awsat"
    mufassal_qisar = "mufassal_qisar"

    def label(self) -> str:
        return {
            PartsNameEnum.as_sab_u_t_tiwal: "Les 7 longues",
            PartsNameEnum.al_miin: "Les Miin",
            PartsNameEnum.al_mathani: "Les Mathani",
            PartsNameEnum.al_mufassal: "Le Mufassal",
            PartsNameEnum.mufassal_tiwal: "Mufassal Long",
            PartsNameEnum.mufassal_awsat: "Mufassal Moyen",
            PartsNameEnum.mufassal_qisar: "Mufassal Court",
        }[self]


class PartsMemoryQuranData(TypedDict):
    as_sab_u_t_tiwal: MemoryQuranData
    al_miin: MemoryQuranData
    al_mathani: MemoryQuranData
    al_mufassal: MemoryQuranData
    mufassal_tiwal: MemoryQuranData
    mufassal_awsat: MemoryQuranData
    mufassal_qisar: MemoryQuranData


class MemoryParts:
    """
    Contient les différentes sections traditionnelles du Coran
    utilisées dans la mémorisation (hifz).
    """

    @property
    def as_sab_u_t_tiwal(self) -> MemoryQuranData:
        """
        Les 7 longues sourates du Coran.
        Traditionnellement, cela inclut les sourates 2 à 9, même si certains savants
        discutent l'inclusion de la sourate 9 (At-Tawbah).
        """
        return MemoryQuranData(start=2, end=9)

    @property
    def al_miin(self) -> MemoryQuranData:
        """
        Les sourates de longueur moyenne entre les longues et les Mathani.
        Sourates 10 à 33.
        """
        return MemoryQuranData(start=10, end=33)

    @property
    def al_mathani(self) -> MemoryQuranData:
        """
        Les sourates "répétées", plus courtes, souvent riches en exhortations.
        Sourates 34 à 49.
        """
        return MemoryQuranData(start=34, end=49)

    @property
    def al_mufassal(self) -> MemoryQuranData:
        """
        Les sourates très souvent récitées par le Prophète ٠.
        Elles sont en général courtes et commencent à la sourate Qaf (50).
        Sourates 50 à 114.
        """
        return MemoryQuranData(start=50, end=114)

    @property
    def mufassal_tiwal(self) -> MemoryQuranData:
        """
        Partie longue du Mufassal.
        Sourates 50 à 77.
        """
        return MemoryQuranData(start=50, end=77)

    @property
    def mufassal_awsat(self) -> MemoryQuranData:
        """
        Partie moyenne du Mufassal.
        Sourates 78 à 92.
        """
        return MemoryQuranData(start=78, end=92)

    @property
    def mufassal_qisar(self) -> MemoryQuranData:
        """
        Partie courte du Mufassal.
        Très souvent mémorisée en premier par les enfants.
        Sourates 93 à 114.
        """
        return MemoryQuranData(start=93, end=114)

    @property
    def all_parts(self) -> PartsMemoryQuranData:
        return {
            "as_sab_u_t_tiwal": self.as_sab_u_t_tiwal,
            "al_miin": self.al_miin,
            "al_mathani": self.al_mathani,
            "al_mufassal": self.al_mufassal,
            "mufassal_tiwal": self.mufassal_tiwal,
            "mufassal_awsat": self.mufassal_awsat,
            "mufassal_qisar": self.mufassal_qisar,
        }


class MemoryQuran:
    def __init__(self) -> None:
        self._quran = MetaSurahs()
        self._memory_parts = MemoryParts()

    def get_parts(self, part: PartsNameEnum, respapi: bool = True):
        memory_parts = self._memory_parts.all_parts[part.value]
        memory_parts_with_surahs = self._quran.get_surahs(
            memory_parts.start, memory_parts.end
        )
        return (
            memory_parts_with_surahs
            if respapi
            else pd.DataFrame(memory_parts_with_surahs)
        )

    def get_all_parts(self, respapi: bool = True) -> Dict[PartsNameEnum, ListMetaSurahs]:
        return {
            part: self.get_parts(part, respapi=respapi)
            for part in PartsNameEnum
        }

    def get_surah_names(self, part: PartsNameEnum) -> List[str]:
        return self.get_parts(part)["name_french"].tolist()

    def get_keywords_learning_methods(self) -> List[str]:
        return [e.value for e in PartsNameEnum]



from .quran_reader_base import QuranReaderBase
from .interface import (
    QuranSourateData,
    QuranVersetData,
    FilterVersets,
    FilterVersetsDict,
)
from .exception_error import SurahNotFound, VersetNotFound, FormatValueGet
from typing import Optional, Dict, List, Tuple, Literal, Union
import re


class MetaQuranReader(QuranReaderBase):
    def __init__(self, path_database):
        super().__init__(path_database)
        self.__lower_name_fr: Optional[Dict[str, QuranSourateData]] = None

    @property
    def lower_name_fr(self) -> Dict[str, QuranSourateData]:
        if self.__lower_name_fr is None:
            self.__lower_name_fr = {
                k.lower(): v for k, v in self.sourate_by_name_fr.items()
            }
        return self.__lower_name_fr

    def get_number_range(self, a: int = 1, b: int = 10) -> List[QuranSourateData]:
        data_quran = self.sourate_by_number
        result: List[QuranSourateData] = []
        for k, v in data_quran.items():
            if a <= k <= b:
                result.append(v)
        return result

    def get_name_fr(self, *names: Tuple[str]) -> Tuple[QuranSourateData]:
        names_lower = [name.lower() for name in list(names)]
        data: List[QuranSourateData] = []
        for name in names_lower:
            sourate = self.lower_name_fr.get(name, None)
            if sourate is None:
                raise SurahNotFound(f"Surah not found {name}")
            data.append(sourate)
        return tuple(data)

    def get_number(self, *numbers: Tuple[int]) -> Tuple[QuranSourateData]:
        data: List[QuranSourateData] = []
        for i in numbers:
            surah = self.sourate_by_number.get(i, None)
            if surah is None:
                raise SurahNotFound(f"Position surah not found {i}")
            data.append(surah)
        return tuple(data)

    def get_arabic(self, *args: Tuple[str]) -> Tuple[QuranSourateData]:
        data: List[QuranSourateData] = []
        for name_arabic in args:
            surah = self.sourate_by_name_arabic.get(name_arabic, None)
            if surah is None:
                raise SurahNotFound(f"Surah not found {name_arabic}")
            data.append(surah)
        return tuple(data)

    def get_simple_page(self, page: int) -> List[QuranSourateData]:
        data_pages: List[QuranSourateData] = []
        for sourate in self.quran["sourates"]:
            for verset in sourate["versets"]:
                if verset["page"] == page and sourate not in data_pages:
                    data_pages.append(sourate)
        return data_pages

    def get_last_page(self) -> List[QuranSourateData]:
        last_surahs = self.search_by_number(114)
        len_versets = len(last_surahs["versets"])
        last_verset = last_surahs["versets"][len_versets - 1]
        last_pages: List[QuranSourateData] = []
        for sourate in self.quran["sourates"]:
            for verset in sourate["versets"]:
                if verset["page"] == last_verset["page"] and sourate not in last_pages:
                    last_pages.append(sourate)
        return last_pages

    def filter_versets(
        self, filter_by: str, value: dict, data: QuranSourateData
    ) -> Optional[QuranVersetData]:
        value_filter = value[filter_by]
        for verset in data["versets"]:
            for k, v in verset.items():
                if filter_by == k and value_filter == v:
                    return verset
        return None

    def get_filter(
        self,
        type_filter: Literal["fr", "arabic", "number"],
        value_filter: Union[str, int],
        type_filter_versets: Optional[FilterVersets] = None,
        value_filter_versets: Optional[FilterVersetsDict] = None,
    ) -> Union[Tuple[QuranSourateData, Optional[QuranVersetData]], Tuple[None, None]]:
        func_dict = {
            "fr": self.get_name_fr,
            "arabic": self.get_arabic,
            "number": self.get_number,
        }
        func_filter = {"position_ds_sourate": self.filter_versets}

        data: Optional[Tuple[QuranSourateData]] = func_dict[type_filter](value_filter)
        if data is None:
            raise SurahNotFound("Surah not found")
        if type_filter_versets is not None:
            data_filter: Optional[QuranVersetData] = func_filter[type_filter_versets](
                type_filter_versets, value_filter_versets, data[0]
            )
            if data_filter is None:
                raise VersetNotFound("Type filter versets not found")
            return data[0], data_filter
        return data[0], None

    def get(self, value: str):
        if not self.check_format_get(value):
            raise FormatValueGet("Format value not matched")
        values = value.split(":")
        if len(values) > 1:
            surah, verset = values
            surah = int(surah)
            verset = int(verset)
            data_s, data_v = self.get_filter(
                "number", surah, "position_ds_sourate", {"position_ds_sourate": verset}
            )
        else:
            values = int(values[0])
            data_s, data_v = self.get_filter("number", values)
        return data_s, data_v

    def check_format_get(self, value: str) -> bool:
        pattern = r"^\d+(?::\d+)?$"
        return bool(re.match(pattern, value))

    def get_versets(
        self,
        surah: "QuranSourateData",
        language: Literal["fr", "arabic"],
        return_types: Literal["str", "list"] = "list",
    ) -> Union[str, List[str]]:
        if language not in ("fr", "arabic"):
            language = "fr"  # Défaut à "fr"

        if return_types == "str":
            results = ""
            for verset in surah["versets"]:
                if language == "fr":
                    results += verset["text"] + "\n"
                else:
                    results += verset["text_arabe"] + "\n"
            return results.strip()
        else:
            results = []
            for verset in surah["versets"]:
                if language == "fr":
                    results.append(verset["text"])
                else:
                    results.append(verset["text_arabe"])
            return results

    


import pandas as pd
import pathlib as plib
from .services import Services
from .quran_reader_base import QuranReaderBase as QuranReader
from .exception_error import ByError, FormatValueGet, SurahNotFound, VersetNotFound
from .interface import ListMetaSurahs
from typing import Literal, Tuple, Union, Optional, List, Any
import re
from langcodes import find as findcodeslg


class ParserMetaSurahs:
    def __init__(self, path_database: str, language: str = "fr") -> None:
        """
        Initialise la classe ParserMetaSurahs qui gère le parsing des métadonnées des sourates.

        Args:
            path_database (str): Le chemin vers la base de données.
        """
        self.path_database = path_database
        self.__services = Services()
        self.__response_api = self._request_response_api()
        self.__quran_reader = QuranReader(path_database)
        self.__language = self.normalize_language_code(language)

    @property
    def language(self) -> str:
        return self.__language

    def normalize_language_code(self, name: str) -> str:
        try:
            lang = findcodeslg(name).language
            return lang if lang != "und" else "en"
        except Exception:
            return "en"

    def _request_response_api(self) -> dict:
        """
        Récupère les données de l'API des métadonnées des sourates.

        Returns:
            dict: La réponse de l'API sous forme de dictionnaire.
        """
        return self.__services.metadata_off_surahs()

    def parser_response_api(self) -> pd.DataFrame:
        """
        Parse les métadonnées des sourates provenant de l'API.

        Returns:
            pd.DataFrame: Un DataFrame contenant les métadonnées des sourates et les informations sur les sajdas.
        """
        responseapi = self.__response_api
        data_response = responseapi["data"]
        surahs = data_response["surahs"]["references"]
        sajdas = data_response["sajdas"]["references"]
        df_surahs = pd.DataFrame(surahs)
        df_sajdas = pd.DataFrame(sajdas)
        df_merged = pd.merge(
            df_surahs, df_sajdas, how="left", left_on=["number"], right_on=["surah"]
        )
        df_merged.drop(columns=["surah", "ayah"], inplace=True)
        df_merged["sajdas_recommended"] = (
            df_merged["recommended"]
            .apply(lambda x: False if pd.isna(x) else x)
            .astype(bool)
        )
        df_merged["sajdas_obligatory"] = (
            df_merged["obligatory"]
            .apply(lambda x: False if pd.isna(x) else x)
            .astype(bool)
        )
        df_merged.drop(columns=["recommended", "obligatory"], inplace=True)
        df_merged.drop_duplicates(inplace=True)
        return df_merged

    def parser_quran_reader(self) -> pd.DataFrame:
        """
        Parse les données des sourates du Coran (sans versets).

        Returns:
            pd.DataFrame: Un DataFrame contenant les informations des sourates (nom, traduction, type de révélation, etc.).
        """
        quran = self.__quran_reader.quran
        sourates_data = [
            {
                "position": sourate["position"],
                "name_arabic": sourate["nom"],
                "name_phonetic": sourate["nom_phonetique"],
                "name_english": sourate["englishNameTranslation"],
                "name_french": sourate["nom_sourate"],
                "revelation_type": sourate["revelation"],
            }
            for sourate in quran["sourates"]
        ]
        df = pd.DataFrame(sourates_data)
        df.drop_duplicates(inplace=True)
        return df

    def parser_meta_surahs(self, save_in: bool = True) -> pd.DataFrame:
        """
        Fusionne les données des sourates provenant de l'API et du fichier du Coran pour créer un DataFrame complet.

        Args:
            save_in (bool): Si True, sauvegarde le DataFrame résultant dans un fichier CSV.

        Returns:
            pd.DataFrame: Un DataFrame contenant toutes les métadonnées des sourates fusionnées.
        """
        parser_quran_reader = self.parser_quran_reader()
        parser_response_api = self.parser_response_api()

        merged_df = pd.merge(
            parser_response_api,
            parser_quran_reader,
            left_on="number",
            right_on="position",
            how="left",
        )

        merged_df = merged_df[
            [
                "number",
                "name_arabic",
                "name_phonetic",
                "name_english",
                "name_french",
                "revelation_type",
                "numberOfAyahs",
                "sajdas_recommended",
                "sajdas_obligatory",
            ]
        ]
        merged_df.rename(
            columns={
                "number": "position",
                "name_arabic": "name_arabic",
                "name_phonetic": "name_phonetic",
                "name_english": "name_english",
                "name_french": "name_french",
                "revelation_type": "revelation_type",
                "numberOfAyahs": "number_of_ayahs",
            },
            inplace=True,
        )
        merged_df["url_quran"] = merged_df["position"].apply(
            lambda x: self.define_url_quran_com(x)
        )
        merged_df.drop_duplicates(inplace=True)
        if save_in:
            path = plib.Path(self.path_database).joinpath(
                f"metasurahs_{self.language}.csv"
            )
            merged_df.to_csv(index=False, path_or_buf=path.as_posix())
        return merged_df

    def define_url_quran_com(self, surah: Any = None) -> str:
        base_url = "https://quran.com/"
        if self.__language == "en":
            return f"{base_url}/{surah}" if surah else base_url
        return (
            f"{base_url}/{self.__language}/{surah}"
            if surah
            else f"{base_url}/{self.__language}"
        )


class MetaSurahs:
    def __init__(self, path_database: str, language: str = "fr") -> None:
        """
        Initialise la classe MetaSurahs en chargeant les données des sourates depuis un fichier CSV
        ou en les générant via le ParserMetaSurahs si le fichier n'existe pas.

        Args:
            path_database (str): Le chemin vers la base de données.
        """
        parser = ParserMetaSurahs(path_database, language)
        lang_code = parser.language
        path = plib.Path(path_database).joinpath(f"metasurahs_{lang_code}.csv")
        if path.exists():
            df = pd.read_csv(path.as_posix())
        else:
            df = parser.parser_meta_surahs()
        self.__df = df
        self.__language = lang_code

    @property
    def language(self) -> str:
        return self.__language

    @property
    def df(self) -> pd.DataFrame:
        """
        Retourne le DataFrame contenant les métadonnées des sourates.

        Returns:
            pd.DataFrame: Le DataFrame des sourates.
        """
        return self.__df

    @property
    def columns_names(self) -> List[str]:
        """
        Retourne la liste des noms de colonnes dans le DataFrame.

        Returns:
            List[str]: La liste des noms de colonnes.
        """
        return list(self.df.columns)

    def get_by(
        self, by: str, value: Union[Any, Tuple, List, None], respapi: bool = True
    ) -> Optional[Union[ListMetaSurahs, pd.DataFrame]]:
        """
        Récupère les enregistrements filtrés par une colonne spécifique et une valeur.
        Si value est None, retourne toutes les données de la colonne 'by'.
        Convertit les valeurs en chaînes de caractères pour éviter les erreurs de typage.

        Args:
            by (str): Le nom de la colonne sur laquelle filtrer.
            value (Union[Any, Tuple, List, None]): La valeur ou les valeurs à filtrer, ou None pour récupérer toutes les données de la colonne.
            respapi (bool): Si True, retourne un dictionnaire JSON.

        Returns:
            Optional[Union[ListMetaSurahs, pd.DataFrame]]: Les données filtrées ou None si aucune donnée.
        """
        if by not in self.columns_names:
            raise ByError(f"Column name invalid {by}")
        if value is None:
            data = self.df[[by]]
        elif isinstance(value, (tuple, list)):
            value = [str(v) for v in value]
            data = self.df[self.df[by].astype(str).isin(value)]
        else:
            value = str(value)  # Convertir la valeur en chaîne de caractères
            data = self.df[self.df[by].astype(str) == value]
        if data.empty:
            return None
        return data.to_dict(orient="records") if respapi else data

    def get_all(self, respapi: bool = True) -> Union[ListMetaSurahs, pd.DataFrame]:
        """
        Retourne toutes les données.

        Args:
            respapi (bool): Si True, retourne les données formatées pour une API (sous forme de liste de dictionnaires).
                            Si False, retourne les données sous forme brute (DataFrame).

        Returns:
            Union[List[dict], pd.DataFrame]: Toutes les données sous forme de dictionnaire (API) ou DataFrame (brut).
        """
        return self.df.to_dict(orient="records") if respapi else self.df

    def sort_by_ayahs(
        self, order: Literal["asc", "desc"], respapi: bool = True
    ) -> Union[ListMetaSurahs, pd.DataFrame]:
        """
        Trie les données par nombre de versets (ayahs).

        Args:
            order (Literal['asc', 'desc']): L'ordre de tri (ascendant ou descendant).
            respapi (bool): Si True, retourne un dictionnaire JSON.

        Returns:
            Union[ListMetaSurahs, pd.DataFrame]: Les données triées.
        """
        data = self.df.sort_values(
            by="number_of_ayahs", ascending=True if order == "asc" else False
        )
        return data.to_dict(orient="records") if respapi else data

    def check_format_get(self, value: str) -> bool:
        pattern = r"^\d+(?::\d+)?$"
        return bool(re.match(pattern, value))

    def _extract_get(self, value: str) -> Tuple[int, Optional[int]]:
        """
        Extrait les valeurs de la requête get.
        """
        if not self.check_format_get(value):
            raise FormatValueGet("Format value not matched")
        values = value.split(":")
        if len(values) > 1:
            surah, verset = values
            surah = int(surah)
            verset = int(verset)
            return surah, verset
        else:
            values = int(values[0])
            return values, None

    def get(self, value: str):
        """
        Récupère les données d'une sourate et d'un verset spécifiques.
        """
        surah, verset = self._extract_get(value)
        data_s = self.get_by("position", surah)
        if data_s is None:
            raise SurahNotFound("Surah not found")
        if verset is not None:
            if verset > data_s[0]["number_of_ayahs"]:
                raise VersetNotFound("Verset not found")
            for surah in data_s:
                new_url = surah["url_quran"] + f"?startingVerse={verset}"
                surah["url_quran"] = new_url
        return data_s

    def get_surahs(self, pos_a: int, pos_b: int) -> ListMetaSurahs:
        """
        Récupère les données des sourates entre deux positions.
        """
        data = self.df[(self.df["position"] >= pos_a) & (self.df["position"] <= pos_b)]
        return data.to_dict(orient="records")

    def get_by_nb_ayahs(self, nb_ayah: int) -> ListMetaSurahs:
        """
        Récupère les sourates ayant un nombre spécifique de versets (ayahs).

        Args:
            nb_ayah (int): Le nombre de versets (ayahs) à rechercher.

        Returns:
            ListMetaSurahs: Une liste de dictionnaires contenant les métadonnées des sourates correspondantes.
        """
        data = self.df[self.df["number_of_ayahs"] == nb_ayah]
        return data.to_dict(orient="records")



from collections import namedtuple


class Config:
    @staticmethod
    def url_meta_data_v1() -> str:
        return "https://api.alquran.cloud/v1/meta"

    @staticmethod
    def url_data_quran() -> str:
        return "https://raw.githubusercontent.com/mehdi-stark/Coran-Quran/refs/heads/master/quran.json"

    @staticmethod
    def author_quran_github() -> str:
        return "https://github.com/mehdi-stark"

    @staticmethod
    def handles_yt():
        return namedtuple(
            "HandlesYT", ["lislamsimplement", "larabesimplement", "lecoransimplement"]
        )(
            larabesimplement="@larabesimplement",
            lecoransimplement="@lecoransimplement",
            lislamsimplement="@lislamsimplement"
        )





import requests
from requests.exceptions import HTTPError, Timeout, RequestException
from .config import Config
from typing import Union


class Services:
    def __init__(self) -> None:
        pass

    def request_services(self, url: str, to_json: bool = False, to_text: bool = False) -> Union[dict, str, requests.Response]:
        try:
            response = requests.get(url, timeout=10)
            response.raise_for_status()

            if to_json:
                return response.json()
            elif to_text:
                return response.text
            else:
                return response

        except HTTPError as http_err:
            raise Exception(f"HTTPError: {http_err}")
        except Timeout:
            raise Exception("Request timeout")
        except RequestException as req_err:
            raise Exception(f"RequestException: {req_err}")
        except Exception as e:
            raise Exception(f"General error: {e}")

    def metadata_off_surahs(self) -> dict:
        return self.request_services(Config.url_meta_data_v1(), to_json=True)

    def quran_to_json(self) -> dict:
        return self.request_services(Config.url_data_quran(), to_json=True)

    def dictionnary_to_dic(self) -> str:
        return self.request_services(Config.dictionnary_fr(), to_text=True)



from typing import Any, NamedTuple, List, Dict, TypedDict
from googleapiclient.discovery import build
from googleapiclient.errors import HttpError
from .config import Config


class Channels(NamedTuple):
    lislamsimplement: str
    larabesimplement: str
    lecoransimplement: str


class NameChannels(NamedTuple):
    lislamsimplement: str
    larabesimplement: str
    lecoransimplement: str


class Playlist(TypedDict):
    title: str
    id: str
    video_count: int
    thumbnail: str
    url_playlist: str
    description: str


class Video(TypedDict):
    title: str
    video_id: str
    position: int
    published_at: str
    thumbnail: str
    url: str


class YoutubeServices:
    def __init__(self, api_key: str) -> None:
        self.api_key = api_key
        self.youtube: Any = build(
            "youtube", "v3", developerKey=api_key, cache_discovery=False
        )
        self.handles = Config.handles_yt()
        self.channels = self.__connect_id_channels()
        self.channel_names = self.__connect_title_channels()

    def __repr__(self) -> str:
        return f"<YoutubeServices connected with key=****{self.api_key[-4:]}>"

    def __get_channel_data(self, handle: str) -> Dict[str, Any] | None:
        try:
            req = self.youtube.search().list(
                part="snippet", type="channel", q=handle, maxResults=1
            )
            res = req.execute()
            return res["items"][0] if res["items"] else None
        except HttpError as e:
            raise e

    def get_channel_id(self, handle: str) -> str | None:
        data = self.__get_channel_data(handle)
        return data["snippet"].get("channelId") if data else None

    def get_name_channel_id(self, handle: str) -> str | None:
        data = self.__get_channel_data(handle)
        return data["snippet"].get("title") if data else None

    def __connect_title_channels(self) -> NameChannels:
        channels = {
            k: self.get_name_channel_id(v) for k, v in self.handles._asdict().items()
        }
        return NameChannels(**channels)

    def __connect_id_channels(self) -> Channels:
        channels = {
            k: self.get_channel_id(v) for k, v in self.handles._asdict().items()
        }
        return Channels(**channels)

    def get_playlist(self, channel_id: str) -> List[Playlist]:
        playlists = []
        next_page_token = None
        while True:
            req = self.youtube.playlists().list(
                part="snippet, contentDetails",
                channelId=channel_id,
                maxResults=50,
                pageToken=next_page_token,
            )
            res = req.execute()
            playlists.extend(res.get("items", []))
            next_page_token = res.get("nextPageToken")
            if not next_page_token:
                break
        return self.__parser_playlist(playlists)

    def __parser_playlist(self, playlists: List[Dict[str, Any]]) -> List[Playlist]:
        data = []
        for playlist in playlists:
            snippet = playlist.get("snippet", {})
            content = playlist.get("contentDetails", {})
            data.append(
                {
                    "title": snippet.get("title", ""),
                    "id": playlist.get("id", ""),
                    "video_count": content.get("itemCount", 0),
                    "thumbnail": snippet.get("thumbnails", {})
                    .get("medium", {})
                    .get("url", ""),
                    "url_playlist": f"https://www.youtube.com/playlist?list={playlist.get('id', '')}",
                    "description": snippet.get("description", ""),
                }
            )
        return data

    def get_videos_playlist(self, playlist_id: str) -> List[Video]:
        videos = []
        next_page_token = None
        while True:
            req = self.youtube.playlistItems().list(
                part="snippet",
                playlistId=playlist_id,
                maxResults=50,
                pageToken=next_page_token,
            )
            res = req.execute()
            videos.extend(res.get("items", []))
            next_page_token = res.get("nextPageToken")
            if not next_page_token:
                break
        return self.__parser_videos(videos, playlist_id)

    def __parser_videos(
        self, videos: List[Dict[str, Any]], playlist_id: str
    ) -> List[Video]:
        data = []
        for video in videos:
            snippet = video.get("snippet", {})
            resource = snippet.get("resourceId", {})
            video_id = resource.get("videoId", "")
            data.append(
                {
                    "title": snippet.get("title", ""),
                    "video_id": video_id,
                    "position": snippet.get("position", 0),
                    "published_at": snippet.get("publishedAt", ""),
                    "thumbnail": snippet.get("thumbnails", {})
                    .get("medium", {})
                    .get("url", ""),
                    "url": f"https://www.youtube.com/watch?v={video_id}&list={playlist_id}",
                }
            )
        return data
